#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#

$_LocalKind = "__LocalKind";
$_ParamKind = "__ParamKind";

sub cfutils_init
{
    $NextTempName = "fmtemp000";
    $NextTempLabel = "12300";
    %UserTypes = ();
    %PortDict = ();
    %PortDictRev = ();
    $NextPortNum = 0;
}

sub applyToPrograms
{
    local($func, @args) = @_;
    local($prog);

    for $prog (@Programs)
    {
	&pushState($prog);
	&$func(@args);
	&popState;
    }
}
    
sub declareSubprogram
{
    local($flavor, $name, $isParallel, $returnType) = @_;
    local($let);

    $CurrentProgram = $name;
    push(@Programs, $name);
    $Programs{$name} = join($;, $flavor, $isParallel, $returnType);

    &setupNewState($name);

#    &processImplicitDecl('real:a-z,integer:i-n');

    for $let ('a' .. 'z')
    {
	$ImplicitTypes{$let} = 'real';
    }
    for $let ('i' .. 'n')
    {
	$ImplicitTypes{$let} = 'integer';
    }
}

sub markSubprogramParallel
{
    local($name) = @_;
    local($flavor, $isParallel, $returnType, $info);

    if (($info = $Programs{$name}) ne '')
    {
	($flavor, $isParallel, $returnType) = split(/$;/, $info);
	$Programs{$name} = join($;, $flavor, 1, $returnType);
    }
}

sub addRawCode
{
    push(@RawCode, @_);
}

sub addDeclCode
{
    push(@DeclCode, @_);
}

sub addInitCode
{
    push(@InitCode, @_);
}

sub addCleanupCode
{
    push(@CleanupCode, @_);
}

sub addBodyCode
{
    push(@BodyCode, @_);
}

 #
 # Declare a common block.
 #
 # Creates the type and dimension assocs for this common block. 
 # 

sub declareCommon
{
    local($commonName) = @_;

    &debug'print("Declare common \"$commonName\"\n");

    $Commons{$commonName} = &newArray;
}

sub declareProgramArg
{
    local($name);

    for $name (@_)
    {
	&debug'print("Declare '$name' as program arg\n");
	push(@Args, $name);
	&vardict'setKind($name, $_ParamKind);
    }
}

sub declareTemp
{
    local($type) = @_;
    local($name);
#    local(@whom) = caller;

    $name = $NextTempName++;

#    print "@whom declares temp $name\n";
    &declareVar($type, $name);

    return $name;
}

sub declareVar
{
    local($type, @names) = @_;
    local($name, $kind);
#    local(@whom) = caller;

#    print "@whom declares $type @names\n";

    for $name (@names)
    {
	if (($kind = &vardict'getKind($name)) eq '')
	{
	    $kind = $_LocalKind;
	    &vardict'setKind($name, $kind);

	    push(@Locals, $name);
	}
	
	&debug'print("declareVar: '$name' '$type' is '$kind'\n");

	&vardict'setType($name, $type);
	&vardict'set($name, 'what','var');
    }
}

sub dimensionVar
{
    local($name, $dim) = @_;
    local($odim);

    &debug'print("dimension '$name' '$dim'\n");

    $odim = &vardict'getDimension($name);

    if ($odim && $dim && $odim ne $dim)
    {
	&warning("Dimension mismatch for $name: $odim $dim");
    }
    &vardict'setDimension($name, $dim);
}

sub declareCommonVar
{
    local($common, @names) = @_;
    local($name, $kind, $l, $i);
    
    for $name (@names)
    {
	$kind = &vardict'getKind($name);

	&debug'print("declareCommonVar '$name' kind was '$kind'\n");

	if ($kind eq $_ParamKind)
	{
	    &error(&vardict'get($name, 'tree_node'),
		   "Cannot declare a parameter to be in a common block");
	}
	elsif ($kind eq $_LocalKind || $kind eq '')
	{
	    $kind = $common;
	    &debug'print("Declaring local $name in common block $common\n");
	    $i = 0;
	    for $l (@Locals)
	    {
		if ($l eq $name)
		{
		    splice(@Locals, $i, 1);
		    last;
		}
		$i++;
	    }
		
	}
	elsif ($kind ne '' && $kind ne $common)
	{
	    &error(&vardict'get($name, 'tree_node'),
		   "Redeclaring common var $name: was in common $kind, now in common $common");
	}

	&debug'print("declareCommonVar '$name' kind is '$kind'\n");
	&vardict'setKind($name, $kind);
	&addVarToCommon($common, $name);
    }
}

sub addVarToCommon
{
    local($common, @vars) = @_;

    if ($Commons{$common} eq '')
    {
	&error("addVarToCommon: undefined common $common\n");
    }
    *array = $Commons{$common};
#    print "addvar: array was '", join(":", @array), "'\n";
    push(@array, @vars);
#    print "addvar: array is now '", join(":", @array), "'\n";
}

sub retrieveVar
{
    local($var) = @_;
    local($kind, $whence);

    $kind = &vardict'getKind($var);
    $type = &vardict'getType($var);
    $dim = &vardict'getDimension($var);

#    print "retrieve gets '$kind' '$type' '$dim'\n";

    return undef if $kind eq '';

    return ($kind, $type, $dim);
}

sub isArg
{
    local($var) = @_;
    local($kind);

    $kind = &vardict'getKind($var);

    return $kind eq $_ParamKind;
}

sub isLocal
{
    local($var) = @_;
    local($kind);

    $kind = &vardict'getKind($var);

    return $kind eq $_LocalKind;
}

sub isParameter
{
    local($var) = @_;

    return $Parameters{$var} ne '';
}

sub isCommon
{
    local($var) = @_;
    local($kind);

    $kind = &vardict'getKind($var);

    return $kind ne '' && $kind ne $_LocalKind && $kind ne $_ParamKind;
}


1;
