#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#

# 
# Write the internal format to a valid fortran input file
#

sub writeFortran
{
    &applyToPrograms("writeFortranProg", $_[0]);
}
    
sub writeFortranProg
{
    local($fh) = @_;
    local($flavor, $isParallel, $returnType) =
	split(/$;/, $Programs{$CurrentProgram});
    local($args, $what);

    local(%varsWritten) = ();
    local($wroteImplicit);
    local($thisname, $needAux, $oname);

    if ($CurrentProgram =~ /\$/ && $flavor eq "subroutine" && $isParallel)
    {
	$needAux = 1;
	($oname = $CurrentProgram) =~ s/\$/__/g;
	$thisname = $oname;
    }
    else
    {
	$thisname = $CurrentProgram;
    }

    print "writeFortranProg $CurrentProgram $oname $thisname\n" if $opt_v;
    
    if ($flavor ne 'program')
    {
	$args = "(" . join(", ", @Args) . ")";
    }
    

    $what = $isParallel ? "parallel" : "sequential";

    print $fh "$commentChar\n";
    print $fh "$commentChar FM transformed $flavor $CurrentProgram\n";
    print $fh "$commentChar This is a $what procedure\n";
    print $fh "$commentChar\n";

    if ($flavor eq 'program' && !$opt_nexus && !$opt_proc)
    {
	&writeLine("      subroutine cfmain $args");
    }
    else
    {
	$returnType = &tree_value(&tree_kid($returnType, 0)) if $returnType;
	&writeLine("      $returnType $flavor $thisname $args");
    }

    &writeDecls;
    &writeDeclCode;
    &writeFortranDecls;
    
    &writeFortranInitCode;
    &writeFortranBodyCode;
    &writeFortranRawCode;
    &writeFortranCleanupCode;
    &writeLine("end");

    if ($needAux )
    {
	&writeLine("$flavor $CurrentProgram $args");
	&writeDecls;
	&writeDeclCode;
	&writeFortranDecls;
	&writeLine("call $oname $args");
	&writeLine("end");
    }
}

sub writeDecls
{
    local($kind, $name, $val);

    for (@Order)
    {
	($kind, $name) = split(/$;/);
#	print "writeDecls: kind=$kind name=$name\n";

	if ($kind eq "parameter")
	{
	    $val = $Parameters{$name};
	    &writeLine("parameter ($name = $val)");
	}
	elsif ($kind eq 'implicit')
	{
	    &writeImplicitDecls;
	}
	elsif ($kind eq "var")
	{
	    &writeVarDecls($name);
	    $varsWritten{$name}++;
	}
	elsif ($kind eq "param")
	{
	    &writeParam($name);
	}
	elsif ($kind eq "external")
	{
	    &writeExternal($name);
	}
	elsif ($kind eq "equivalence")
	{
	    &writeLine($name);
	}
    }
}

sub writeParam
{
    local($name) = @_;
    local($val, $str);

    $val = $Parameters{$name};
    $str = &expr_to_string($val);
    print $fh "      parameter ($name = $str)\n";
}
   
sub writeExternal
{
    local($name) = @_;

    print $fh "      external $name\n";
}

sub writeFortranDecls
{
    local($var, $type, $dim, $common);

    &writeVarDecls(@Args, @Locals);
    for $common (keys %Commons)
    {
	local($head, $vars);

	local(*array) = $Commons{$common};
	local(@copy) = @array;
	
	$common = '' if $common eq '<<BLANK>>';
	    
	$head = "      COMMON /$common/";
	
	$vars = '';
	while (@copy)
	{
	    $var = shift(@copy);
	    if ((length($vars) + length($var) + 2) > (70 - length($head)))
	    {
		print $fh "$head $vars\n";
		$vars = '';
	    }
	    $vars .= ", " if $vars;
	    $vars .= $var;
	}
	if ($vars ne '')
	{
	    print $fh "$head $vars\n";
	}

#	print "writing var decls for common vars '@array'\n";
	&writeVarDecls(@array);
    }
}

sub writeImplicitDecls
{
    local($string);
    local($onedecl, $type, $letters, $lsets, $letter);

    return if $wroteImplicit;
    $wroteImplicit = 1;

    if ($NoImplicitTypes)
    {
	print $fh "      implicit none\n";
    }
    else
    {
	for $string (@ImplicitStatementList)
	{
	    
	    for $onedecl (split(/,/, $string))
	    {
		($type, $letters) = split(/:/, $onedecl);
		
		$type =~ s/_/ /g;
		
		$letters =~ s/\+/,/g;
		
		print $fh "      implicit $type ($letters)\n";
	    }
	}
    }
}    

sub writeVarDecls
{
    local($var, $type, $dim);

    for $var (@_)
    {
	next if $varsWritten{$var};
	$type = &vardict'getType($var);
	if ($type ne '')
	{
	    if ($type eq 'character')
	    {
		$dim = &vardict'get($var, 'character_size');
		print $fh "      character*$dim $var\n";
	    }
	    else
	    {
		$type =~ s/_/ /g;
		print $fh "      $type $var\n";
	    }
	}
    }

    for $var (@_)
    {
	next if $varsWritten{$var};
	$dim = &vardict'getDimension($var);
	if ($dim ne '')
	{
	    if ($type eq 'character')
	    {
		$dim =~ s/^[^|]+(\|?)//;
	    }
	    $dim =~ s/\|/, /g;
	    print $fh "      DIMENSION $var($dim)\n" if $dim ne '';
	}
    }
}

sub writeFortranInitCode
{
    local($line);

    print $fh "$commentChar\n$commentChar Beginning of initial code\n$commentChar\n";
    for $line (@InitCode)
    {
	&writeLine($line);
    }
}

sub writeDeclCode
{
    local($line);

    print $fh "$commentChar\n$commentChar Beginning of decl code\n$commentChar\n";
    for $line (@DeclCode)
    {
	&writeLine($line);
    }
}

sub writeFortranCleanupCode
{
    local($line);

    print $fh "$commentChar\n$commentChar Beginning of cleanup code\n$commentChar\n";
    for $line (@CleanupCode)
    {
	&writeLine($line);
    }
}

sub writeFortranBodyCode
{
    local($line);

    print $fh "$commentChar\n$commentChar Beginning of body code\n$commentChar\n";
    for $_ (@BodyCode)
    {
	($line) = &hackLine($_);
	&writeLine($line);
    }
}

sub writeFortranRawCode
{
    local($line);

    print $fh "$commentChar\n$commentChar Beginning of raw code\n$commentChar\n";
    for $line (@RawCode)
    {
	&writeLine($line);
    }
}

sub writeLine
{
    local($line) = @_;
    
    print "<$line>\n" if $ENV{"SHOW_LINES"};

    if ($line =~ /^%/)
    {
	$line =~ s/^%/$commentChar     /;
	print $fh "$line\n";
    }
    else
    {
	if (substr($line, 0, 5) =~ /^\s*(\d+)/)
	{
	    print $fh "$1\t";
	    $line = substr($line, 5);
	}
	else
	{
	    print $fh "\t";
	}
	
	$needcont = 0;
	

	print $fh "$line\n";
	return;
	
	while (length($line) > 60)
	{
	    $i = -1;
	    for ($start = 50; $i < 0 && $start >= 0; $start -= 10)
	    {
		$i = index($line, " ", $start);
	    } 
	    
	    $i = length($line) if $i < 0;
	    $first = substr($line, 0, $i);
	    $line = substr($line, $i);
#		print "i=$i first=<$first> line=<$line>\n";
	    if ($needcont)
	    {
		print $fh "     \$ ";
	    }
	    print $fh $first, "\n";
	    $needcont = 1;
	}
	if ($needcont)
	{
	    print $fh "     \$ ";
	}
	print $fh $line, "\n";
    }
}    

    

1;
