#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
#
# Generate the code for spawn_procedure_()
#
    
sub genSpawnCode
{
    local($fh) = @_;
    local($flavor, $isParallel);
    local($name, @args);
    local($nargs);
    local($i, $arg, $ctype);
    local(@argtypes, @argdims, @argnames, @argintents, @argdimstrs, @amper);
    local(@argStar, @argStructType);
    local(@charArgs, $charLenDecls, $localCharLenDecls, $charLengths);
    local($localCharVarDecls, $extractCharVars, $extractCharLengths);
    local(@charVar, @charLen);
    local(@inports, @outports, @nonports);
    local(@outargs, @inargs);
    local(@portnames);
    local(@argIn, @argOut);
    local(@argGP);
    local($ptrs_struct, $storage_struct);
    local($str);
    local($startFuncId, $startFuncName);
    local($copyProcName, $setupProcName, $spawnProcName);
    local($packProcName, $argsStructName, $outargsStructName);
    local($spawnProcNameWithoutTrailing, $spawnProcNameWithTrailing);
    local($setupProcNameWithoutTrailing, $setupProcNameWithTrailing);
    local($callName);

    ($flavor, $isParallel) = split(/$;/, $Programs{$CurrentProgram});

    if (!$isParallel)
    {
#	print "genSpawn: skipping $CurrentProgram\n";
	return;
    }

    $name = $CurrentProgram;

    $name =~ s/\$/__/g;

    $nargs = @Args;

    $startFuncId = "fmu_${name}_id";
    $startFuncName = "fmu_remote_start_${name}";
    $copyProcName = "fmu_copyfunc_${name}";
    $setupProcName = "FMU_T\U${name}$nargs";
#    $setupProcName = "FMU_SETUP_\U${name}_$nargs";
    $packProcName = "fmu_packargs_${name}";
    $spawnProcName = "FMU_P\U${name}$nargs";
#    $spawnProcName = "FMU_SPAWN_\U${name}_$nargs";
    $argsStructName = "fmt_${name}_args";
    $outargsStructName = "fmt_${name}_outargs";

    push(@FM_Handlers, $startFuncName . $; . $startFuncId . $; . $CurrentProgram);

    $callName = "\U${name}";

    print $fh "#ifdef FM_NEED_TRAILING_UNDERSCORES\n\n";
    print $fh "#define $setupProcName \L${setupProcName}_\n";
    print $fh "#define $spawnProcName \L${spawnProcName}_\n";
    print $fh "#define $callName ${name}_\n";
    print $fh "\n#else /* FM_NEED_TRAILING_UNDERSCORES */\n";
    print $fh "\n#ifndef FM_NEED_UPPERCASE\n\n";
    print $fh "#define $setupProcName \L${setupProcName}\n";
    print $fh "#define $spawnProcName \L${spawnProcName}\n";
    print $fh "#define $callName ${name}\n";
    print $fh "\n#endif /* FM_NEED_UPPERCASE */\n";
    print $fh "\n#endif /* FM_NEED_TRAILING_UNDERSCORES */\n";

    print $fh (<<END);

extern fm_remote_start_func_id_t $startFuncId;

extern void ${callName}();
    
END


    &analyzeArgs;

    if ($opt_v)
    {
	for ($i = 0; $i < $nargs; $i++)
	{
	    print "Argument $i: <$argnames[$i]> <$argdims[$i]> <$argdimstrs[$i]>, <$argtypes[$i]>\n";
	}
    }

    &genOutargsStruct;
    &genArgsStruct;
    &genCopyProc;
    &genSetupProc;
    &genPackProc;
    &genSpawnProc;
    &genStartProc;
}

sub analyzeArgs
{
    local($i);
    local($implicitIntent);
    local($intent, $dimstr, $tree, $dimnum);

    $i = 0;
    for $arg (@Args)
    {
	$implicitIntent = 0;
	$type = &vardict'getOriginalType($arg);
	$type = &vardict'getType($arg) if $type eq '';
	$dim = &vardict'getDimension($arg);

	if ($dim)
	{
	    local($adim, $cdim, @dimExprs, @dimNames);
	    $cdim = 1;

	    @dimExprs = &grokDims($arg);
	    @dimNames = split(/\|/, $dim);

#	    print "Got dimExprs <", join(":", @dimExprs), "> arg=<$arg>\n";
#	    print "Got dim <$dim>\n";

	    $dimnum = 0;

#	    for $adim (split(/\|/, $dim))
	    for $adim (@dimExprs)
	    {
		$dimnum++;
		if ($adim =~ /^\d+$/)
		{
		    $cdim *= $adim;
		}
		else
		{
		    local($num) = $i + 1;
		    &error($tree,
			   "process $name argument $num $arg: invalid subscript $dimnum ($dimNames[$i])");
		}
	    }
	    $dimstr = "$cdim";
	}
	else
	{
	    $dimstr = '';
	}

	if ($type eq '')
	{
	    local($argno) = $i + 1;
	    &error($ProgramTree{$CurrentProgram},
		   "Argument $argno ($arg) of $CurrentProgram is undefined");

	}

	$isPortVar[$i] = &isPortType($type);

	if ($type eq "in_port")
	{
	    push(@inports, $i);
	}
	elsif ($type eq "out_port")
	{
	    push(@outports, $i);
	}
	else
	{
	    push(@nonports, $i);
	}

	$type = "f_$type";

	$argname = "A_$i";

	$intent = &vardict'get($arg, 'intent');
	if ($intent eq '')
	{
	    $implicitIntent++;
#	    &warning(&vardict'get($arg, 'tree_node'),
#		     "Implicit intent declaration of $arg")
#		unless $isPortVar[$i];
	    $intent = 'inout';
	}

	
	if (($intent eq 'in' || $intent eq 'inout') && !$isPortVar[$i])
	{
	    push(@inargs, $i);
	    $argIn[$i] = 1;
	}

	if ($intent eq 'out' || $intent eq 'inout')
	{
	    if ($isPortVar[$i])
	    {
		if (!$implicitIntent)
		{
		    &error(&vardict'get($arg, 'tree_node'),
			   "A port cannot be an output argument");
		    undef $argOut[$i];
		}
	    }
	    else
	    {
		$argOut[$i] = 1;
		push(@outargs, $i);
	    }
	}

	    
	$argnames[$i] = $argname;
	$argdims[$i] = $dim;
	$argdimstrs[$i] = $dimstr;
	$argdimsizes[$i] = $dimstr eq '' ? 1 : $dimstr;
	$amper[$i] = $dimstr eq '' ? '&' : '';
	$argtypes[$i] = $type;
	if ($type eq 'f_character')
	{
	    push(@charArgs, $i);
	    $argStar[$i] = '';
	    $argStructType[$i] = 'char';
	    $charVar[$i] = "$argnames[$i]_charvar";
	    $charLen[$i] = "$argnames[$i]_charlen";
	}
	else
	{
	    $argStar[$i] = '*';
	    $argStructType[$i] = $type;
	}
	$argrealtypes[$i] = $isPortVar[$i] ? "f_integer" : $argtypes[$i];
	$i++;
    }

    $charLengths = '';
    $charLenDecls = '';
    $localCharLenDecls = '';
    $localCharVarDecls = '';
    $extractCharLengths = '';
    $extractCharVars = '';
    for $i (@charArgs)
    {
	$charLengths .= "CHAR_LEN($charLen[$i]) ";
	$charLenDecls .= "\tCHAR_LEN_DECL($charLen[$i])\n";
	$localCharLenDecls .= "    LOCAL_CHAR_LEN_DECL($charLen[$i])\n";
	$localCharVarDecls .= "    char *$charVar[$i];\n";
	$extractCharVars .= "    EXTRACT_CHAR_PTR($argnames[$i], $charVar[$i])\n";
	$extractCharLengths .= "    EXTRACT_CHAR_LEN($argnames[$i], $charLen[$i])\n";
    }
}

#
# grokDims()
#
# Scan the dimensions of this variable. Return a list of dimension
# information items, each consisting of
#
#     the size of the dimension, if integral; or -1 if not
#     an expression in terms of A<argnum> specifying the size of a variable
#        size dimension
#
# The above is for some future implementation; for now I'll only
# support constant-sized array arguments.
#
sub grokDims
{
    local($arg) = @_;
    local($tree, $var, $type_name, $elt, $l, $u, $csize);
    local(@dims);
    
    $tree = &vardict'get($arg, 'tree_node');
    $var = &tree_kid(&tree_kid($tree, 1), 0);

#    &dump_parse_tree($tree);
#    &dump_parse_tree($var);

    $type_name = &tree_type_name($var);

#    print "grok: $arg $type_name\n";

    if ($type_name eq "string")
    {
	#
	# No array dimension here.
	#
	if (($csize = &vardict'get($arg, 'character_size')))
	{
	    ($csize);
	}
	else
	{
	    undef;
	}
    }
    elsif ($type_name eq "array_declarator")
    {
	$bounds = &tree_kid($var, 1);

	for $elt (&expand_tree_list($bounds))
	{
	    $type_name = &tree_type_name($elt);
#	    print "Got bounds elt type $type_name\n";
	    if ($type_name eq "lower_upper")
	    {
		$l = &tree_kid($elt, 0);
		$u = &tree_kid($elt, 1);

		$l = &evaluateExpr($l);
		$u = &evaluateExpr($u);

		push(@dims, $l . $; . $u);
	    }
	    else
	    {
		$l = &evaluateExpr($elt);
		if (!$l)
		{
#		    &error($elt, "Invalid dimension");
		}
		push(@dims, $l);
	    }
	}
	if (($csize = &vardict'get($arg, 'character_size')))
	{
#	    print "Grok adds $csize to dims for $arg\n";
	    unshift(@dims, $csize);
	}
	@dims;
    }
    else
    {
	undef;
    }
}

sub genOutargsStruct
{
    local($i, $star);

    print $fh "struct $outargsStructName\n";
    print $fh "{\n";

    if (@outargs == 0)
    {
	print $fh "    int dummy;\n";
    }
    else
    {
	for $i (@outargs)
	{
	    print $fh "    $argtypes[$i] $argStar[$i]$argnames[$i];\n";
	}
    }
    print $fh "};\n\n";
}

sub genArgsStruct
{
    local($i, $ptype);

    return if $nargs == 0;

    print $fh "struct $argsStructName\n";
    print $fh "{\n";

    for ($i = 0; $i < $nargs; $i++)
    {
	if ($isPortVar[$i])
	{
	    print $fh "    f_integer $argnames[$i]\[$argdimsizes[$i]\];\n";
	    $ptype = $argtypes[$i] eq "f_in_port" ? "fm_inport_t" : "fm_outport_t";
	    print $fh "    $ptype $argnames[$i]_val\[$argdimsizes[$i]\];\n";
	}
	elsif ($argdimstrs[$i])
	{
	    print $fh "    $argStructType[$i] $argnames[$i][$argdimstrs[$i]];\n";
	}
	else
	{
	    print $fh "    $argStructType[$i] $argnames[$i];\n";
	}
    }
    print $fh "};\n\n";
}

sub genCopyProc
{
    local($i);
    local($var);

    print $fh "void $copyProcName(void *arg_list, yoke_buffer_t return_buf)\n";
    print $fh "{\n";

    if (@outargs)
    {
	for $i (@charArgs)
	{
	    if ($argOut[$i])
	    {
		print $fh "    char *$charVar[$i];\n";
	    }
	}
	print $fh "    struct $outargsStructName *args = arg_list;\n\n";

	for $i (@outargs)
	{
	    if ($charVar[$i])
	    {
		print $fh "    EXTRACT_CHAR_PTR(args->$argnames[$i], $charVar[$i])\n";
		$var = $charVar[$i];
	    }
	    else
	    {
		$var = "args->$argnames[$i]";
	    }
	    print $fh "    yoke_get_$argtypes[$i](return_buf, $var, $argdimsizes[$i]);\n";
	}
    }
    print $fh "}\n\n";
}

sub genSetupProc
{
    local($s);
    local($nips, $nops);
    local($j, $i);

    $s = "void $setupProcName(";
    print $fh "$s";
    $s =~ s/./ /g;
    for ($i = 0; $i < $nargs; $i++)
    {
	print $fh $s if $i > 0;
	print $fh "$argrealtypes[$i] $argStar[$i]$argnames[$i],\n";
    }
    print $fh "${s}f_integer *locn,\n";
    print $fh "${s}f_integer *locn_size,\n";
    print $fh "${s}f_integer *f_pblock\n";
    print $fh "$charLenDecls)\n";

    for ($nips = 0, $i = 0; $i < @inports; $i++)
    {
#        print "inport $i size=$argdimsizes[$inports[$i]]\n";
        $nips += $argdimsizes[$inports[$i]];
    }
    for ($nops = 0, $i = 0; $i < @outports; $i++)
    {
#        print "outport $i size=$argdimsizes[$outports[$i]]\n";
        $nops += $argdimsizes[$outports[$i]];
    }

    print $fh (<<END);
{
    fm_process_block_t pblock;
    fm_inport_t *ips;
    fm_outport_t *ops;
    int idx, i;
$localCharLenDecls
    pblock = (fm_process_block_t) *f_pblock;

END

    if ($nips > 0)
    {
	print $fh (<<END);
    ips = (fm_inport_t *) malloc(sizeof(fm_inport_t) * $nips);
    if (ips == (fm_inport_t *) NULL)
	yoke_fatal("$setupProcName(): malloc failed\\n");
END
    }
    else
    {
	print $fh "    ips = (fm_inport_t *) NULL;\n";
    }
    
    if ($nops > 0)
    {
	print $fh (<<END);
    ops = (fm_outport_t *) malloc(sizeof(fm_outport_t) * $nops);
    if (ops == (fm_outport_t *) NULL)
	yoke_fatal("$setupProcName(): malloc failed\\n");
END
    }
    else
    {
	print $fh "    ops = (fm_outport_t *) NULL;\n";
    }
    

    print $fh "    idx = 0;\n";
    for ($j = 0; $j < @inports; $j++)
    {
	$i = $inports[$j];
        print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
        Assert(idx < $nips);
        ips\[idx\] = (fm_inport_t) $argnames[$i]\[i\];
	ips\[idx\]->passed_in_call = FALSE;
	idx++;
    }
END
    }
    print $fh "    idx = 0;\n";
    for ($j = 0; $j < @outports; $j++)
    {
	$i = $outports[$j];
        print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
        Assert(idx < $nops);
        ops\[idx\] = (fm_outport_t) $argnames[$i]\[i\];
	ops\[idx\]->passed_in_call = FALSE;
	idx++;
    }
END
    }

    print $fh (<<END);
    _fm_setup_generic($startFuncId, $copyProcName,
                      ips, $nips, ops, $nops,
                      locn, *locn_size, pblock);
    pblock->current_slot++;
}
END
    
}

sub genPackProc
{
    local($s);
    local($nips, $nops);
    local($j, $i);

    $s = "yoke_buffer_t $packProcName(";
    print $fh "$s";
    $s =~ s/./ /g;
    for ($i = 0; $i < $nargs; $i++)
    {
	print $fh $s if $i > 0;
	print $fh "$argrealtypes[$i] $argStar[$i]$argnames[$i],\n";
    }
    print $fh "${s}struct $outargsStructName **outargs,\n";
    print $fh "${s}fm_process_block_t pblock\n";
    print $fh "$charLenDecls)\n";

    print $fh (<<END);
{
    yoke_buffer_t buf;
    int slot;
    fm_process_block_entry_t *ent;
    conn_t dest;
    int size, count, count_1, i;
$localCharVarDecls
$localCharLenDecls
END

    print $fh "\n";

    for $i (@inports)
    {
	print $fh "    fm_inport_t *$argnames[$i]_port;\n";
    }
    for $i (@outports)
    {
	print $fh "    fm_outport_t *$argnames[$i]_port;\n";
    }

    print $fh "\n";

    for $i (@inports)
    {
	print $fh "    $argnames[$i]_port = (fm_inport_t *) $argnames[$i];\n";
    }
    for $i (@outports)
    {
	print $fh "    $argnames[$i]_port = (fm_outport_t *) $argnames[$i];\n";
    }

    print $fh (<<END);

$extractCharVars
$extractCharLengths

    slot = pblock->current_slot;
    ent = &pblock->entries[slot];
    dest = ent->process;

    size = yoke_sizeof_int(dest, 1);
    count = 1;

END
    for $i (@inports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	size += _fm_sizeof_inport(dest, $argnames[$i]_port\[i\], &count_1);
	count += count_1;
    }
END
    }
    for $i (@outports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	size += _fm_sizeof_outport(dest, $argnames[$i]_port\[i\], &count_1);
	count += count_1;
    }
END
    }
    for $i (@inargs)
    {
	print $fh "    size += yoke_sizeof_$argtypes[$i](dest, $argdimsizes[$i]);\n";
	print $fh "    count++;\n";
	#
	# Why ever did I write this??
	#
#	print $fh "    count += $argdimsizes[$i];\n";
    }

    print $fh "    buf = yoke_allocate_buffer(dest, size, count);\n";

    print $fh "    yoke_put_int(buf, &ent->n_inport_migrations, 1);\n";

    for $i (@inports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	_fm_put_inport(buf, $argnames[$i]_port\[i\]);
    }
END
    }
    for $i (@outports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	_fm_put_outport(buf, $argnames[$i]_port\[i\]);
    }
END
    }
    for $i (@inargs)
    {
	if ($charVar[$i])
	{
	    print $fh "    yoke_put_$argtypes[$i](buf, $charVar[$i], $argdimsizes[$i]);\n";
	}
	else
	{
	    print $fh "    yoke_put_$argtypes[$i](buf, $argnames[$i], $argdimsizes[$i]);\n";
	}
    }
    print $fh (<<END);
    *outargs = (struct $outargsStructName *)
               	       malloc(sizeof(struct $outargsStructName));
    if (*outargs == (struct $outargsStructName *) NULL)
	yoke_fatal("$packProcName(): malloc failed\\n");
END
    
    for $i (@outargs)
    {
	print $fh "    (*outargs)->$argnames[$i] = $argnames[$i];\n";
    }

    print $fh (<<END);

    return buf;
}
END
    
}

sub genSpawnProc
{
    local($s);
    local($j, $i);
    local($args);

    $s = "void $spawnProcName(";
    print $fh "$s";
    $s =~ s/./ /g;
    for ($i = 0; $i < $nargs; $i++)
    {
	print $fh $s if $i > 0;
	print $fh "$argrealtypes[$i] $argStar[$i]$argnames[$i],\n";
    }
    print $fh "${s}f_integer *f_pblock\n";
    print $fh "$charLenDecls)\n";

    $args = join(", ", @argnames, "&outargs", "pblock $charLengths");

    print $fh (<<END);
{
    fm_process_block_t pblock;
    yoke_buffer_t buf;
    struct $outargsStructName *outargs;
$localCharVarDecls
$localCharLenDecls
    pblock = (fm_process_block_t) *f_pblock;

    /*
    _fm_fillin_outports(pblock);
    */
    
    buf = $packProcName($args);

    _fm_spawn_generic(buf, outargs, pblock);
    pblock->current_slot++;
}

END
}

sub genStartProc
{
    local($i, $args);
    
    print $fh (<<END);
void $startFuncName(conn_t _fmint_sender, yoke_buffer_t buf)
{
    struct $argsStructName *args;
    int size, count, i, n_inport_migrations;
END

    for $i (@charArgs)
    {
	print $fh "    f_character $argnames[$i];\n";
	print $fh "    int $charLen[$i];\n";
	print $fh "    char *$charVar[$i];\n";
    }

    if ($nargs > 0)
    {
	print $fh (<<END);
    args = (struct $argsStructName *) malloc(sizeof(struct $argsStructName));
    ZeroOutMemory(args, sizeof(struct $argsStructName));
END
    }

    print $fh (<<END);
    yoke_get_int(buf, &n_inport_migrations, 1);
END
    
    for $i (@inports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	_fm_get_inport(buf, &(args->$argnames[$i]_val\[i\]));
	_fm_pindex_add_inport(args->$argnames[$i]_val\[i\]);
	args->$argnames[$i]\[i\] = (f_integer) args->$argnames[$i]_val\[i\];
    }
END
    }
    for $i (@outports)
    {
	print $fh (<<END);
    for (i = 0; i < $argdimsizes[$i]; i++)
    {
	_fm_get_outport(buf, &(args->$argnames[$i]_val\[i\]));
	_fm_pindex_add_outport(args->$argnames[$i]_val\[i\]);
	args->$argnames[$i]\[i\] = (f_integer) args->$argnames[$i]_val\[i\];
    }
END
    }
    for $i (@inargs)
    {
	print $fh "    yoke_get_$argtypes[$i](buf, $amper[$i]args->$argnames[$i], $argdimsizes[$i]);\n";
    }

#    $args = '&args->' . join(', &args->', @argnames) if @argnames;

    $args = '';
    for ($i = 0; $i < $nargs; $i++)
    {
	if ($charVar[$i])
	{
	    print $fh "    CONVERT_CHAR_C_TO_FTN($argnames[$i], args->$argnames[$i], $charLen[$i])\n";
	    $args .= "$argnames[$i]";
	}
	else
	{
	    $args .= "&args->$argnames[$i]";
	}
	$args .= ", " if $i < $nargs - 1;
    }
    for $i (@charArgs)
    {
	$args .= " CHAR_LEN($charLen[$i])";
    }

    print $fh "    _fm_handle_incoming_migrations(_fmint_sender, n_inport_migrations);\n";
    print $fh "    _fm_handle_queued_migration_requests();\n";
    print $fh "    yoke_free_buffer(buf);\n";
    print $fh "    EndSafeRegion;\n";
    
    print $fh "    ${callName}($args);\n";
    print $fh "    StartSafeRegion;\n";
    print $fh "    size = yoke_sizeof_int(_fm_process_creator, 2);\n";
    print $fh "    count = 2;\n";
    for $i (@outargs)
    {
	print $fh "    size += yoke_sizeof_$argtypes[$i](_fm_process_creator, $argdimsizes[$i]);\n";
	print $fh "    count++;\n";
    }
    
    print $fh (<<END);

    buf = yoke_allocate_buffer(_fm_process_creator, size, count);

#ifdef DEBUG
    if (FMDebug(1))
    {
	yoke_printf(\"Process %s returning to %s, slot=%d\\n\",
		    yoke_conn_as_string(yoke_self()),
		    yoke_conn_as_string(_fm_process_creator),
		    _fm_process_slot);
    }
#endif

    yoke_put_int(buf, &_fm_process_remote_pblock, 1);
    yoke_put_int(buf, &_fm_process_slot, 1);
END

    for $i (@outargs)
    {
	if ($charVar[$i])
	{
	    print $fh "    EXTRACT_CHAR_PTR($argnames[$i], $charVar[$i])\n";
	    print $fh "    yoke_put_char(buf, $charVar[$i], $argdimsizes[$i]);\n";
	}
	else
	{
	    print $fh "    yoke_put_$argtypes[$i](buf, $amper[$i]args->$argnames[$i], $argdimsizes[$i]);\n";
	}
    }
    
    print $fh (<<END);

    yoke_establish(_fm_process_creator);
    yoke_send(_fm_process_creator, buf, FM_PROCESS_RETURN);
    yoke_release(_fm_process_creator);

}

END

}

#
# Generate the handler table for the linker from the array
# @FM_Handlers which consists of pairs Name . $; . Id.
#
sub generateHandlerTable
{
    local($fh) = @_;
    local($name, $id, $process);

    print $fh "fm_remote_start_func_registry_t _fmr_${FileBaseNoSpecials}[] = {\n";

    for (@FM_Handlers)
    {
	($name, $id, $process) = split(/$;/);

	print $fh (<<END)
    { &$id, $name,
      \"$name\",
      \"$process\",
    },
END
    }
    print $fh "    { NULL, NULL },\n";
    print $fh "};\n\n";

    print $fh "/*\n";
    print $fh "NEXUS_HANDLERS_START\n";
    for (@FM_Handlers)
    {
	($name, $id) = split(/$;/);

	print $fh "$name $id\n";
    }
    print $fh "NEXUS_HANDLERS_END\n";
    print $fh "*/\n";
    
}

1;
