#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
#
# Generate wrappers for port send/recv from the information in the portdict.
# 


sub genSRWrappers
{
    local($fh) = @_;

    for (sort keys %PortDict)
    {
	&genSRWrapper($fh, $_);
    }
}

sub genSRWrapper
{
    local($fh, $port) = @_;
    local($portidx, $structname, $msgbuf);
    local($n_elts, @elt_id, @elt_type, @elt_is_array, @elt_ctype, $i);
    local(@elt_is_port);
    local($elt_id, $elt_type, $elt_is_array, $elt_ctype, $elt_is_port);
    local(@array_elts, @nonarray_elts, $n_array_elts);
    local(@port_elts, @port_type, @inport_elts, @outport_elts);
    local($have_port_arrays);
    local($charLengths, $charLenDecls, $localCharLenDecls);
    local($localCharVarDecls, $extractCharLengths, $extractCharVars);
    local(@charVar, @charElts, @charLen);
    
    if ($port =~ /^!port_(\d+)$/)
    {
	$portidx = $1;
    }
    else
    {
	&fatal("Ack! Invalid port $port");
    }

    $n_elts = &portdict'get($port, 'n_elts');

    for ($i = 0; $i < $n_elts; $i++)
    {
	$elt_id = "A_$i";

	$elt_type = &portdict'get($port, $elt_id);
	$elt_is_array = &portdict'get($port, $elt_id . "_is_array");
        $elt_is_port = &portdict'get($port, $elt_id . "_port_ptr");

	$elt_ctype = "f_$elt_type";
	print "elt $i type=$elt_type is_array=$elt_is_array is_port=$elt_is_port\n" if $opt_v;
	if ($elt_is_array)
	{
	    push(@array_elts, $i);
	}
	else
	{
	    push(@nonarray_elts, $i);
	}
        if ($elt_is_port)
        {
            push(@port_elts, $i);
            $elt_ctype = "f_integer";
            $elt_type =~ s/_//;
            $port_type[$i] = "fm_${elt_type}_t";
            if ($elt_type eq "inport")
            {
                push(@inport_elts, $i);
            }
            else
            {
                push(@outport_elts, $i);
            }
                
        }
	if ($elt_type eq "character")
	{
	    $elt_star[$i] = '';
	    push(@charElts, $i);
	    $charVar[$i] = "${elt_id}_charvar";
	    $charLen[$i] = "${elt_id}_charlen";
	}
	else
	{
	    $elt_star[$i] = '*';
	}
	$elt_id[$i] = $elt_id;
	$elt_type[$i] = $elt_type;
	$elt_ctype[$i] = $elt_ctype;
	$elt_is_array[$i] = $elt_is_array;
        $elt_is_port[$i] = $elt_is_port;
        $elt_array_len[$i] = $elt_is_array ? "*$elt_id[$i]_size" : "1";
        $have_port_arrays++ if $elt_is_array && $elt_is_port;
    }
    $charLengths = '';
    $charLenDecls = '';
    $localCharLenDecls = '';
    $localCharVarDecls = '';
    $extractCharLengths = '';
    $extractCharVars = '';
    for $i (@charElts)
    {
	$charLengths .= "CHAR_LEN($charLen[$i]) ";
	$charLenDecls .= "\tCHAR_LEN_DECL($charLen[$i])\n";
	$localCharLenDecls .= "    LOCAL_CHAR_LEN_DECL($charLen[$i])\n";
	$localCharVarDecls .= "    char *$charVar[$i];\n";
	$extractCharVars .= "    EXTRACT_CHAR_PTR($elt_id[$i], $charVar[$i])\n";
	$extractCharLengths .= "    EXTRACT_CHAR_LEN($elt_id[$i], $charLen[$i])\n";
    }

    $n_array_elts = @array_elts;

    $recvName = "\Ufmu_r${FileBaseNoSpecials}${portidx}";
    $sendName = "\Ufmu_s${FileBaseNoSpecials}${portidx}";
    $recvNameDebug = "\Ufmu_r${FileBaseNoSpecials}${portidx}d";
    $sendNameDebug = "\Ufmu_s${FileBaseNoSpecials}${portidx}d";

#    $recvName = "\Ufm_recv_${FileBaseNoSpecials}_port${portidx}";
#    $sendName = "\Ufm_send_${FileBaseNoSpecials}_port${portidx}";
#    $recvNameDebug = "\Ufm_recv_${FileBaseNoSpecials}_port${portidx}_debug";
#    $sendNameDebug = "\Ufm_send_${FileBaseNoSpecials}_port${portidx}_debug";

    print $fh "#ifdef FM_NEED_TRAILING_UNDERSCORES\n\n";
    print $fh "#define $sendName \L${sendName}_\n";
    print $fh "#define $recvName \L${recvName}_\n";
    print $fh "#define $sendNameDebug \L${sendNameDebug}_\n" if $opt_g;
    print $fh "#define $recvNameDebug \L${recvNameDebug}_\n" if $opt_g;
    print $fh "\n#else /* FM_NEED_TRAILING_UNDERSCORES */\n";
    print $fh "\n#ifndef FM_NEED_UPPERCASE\n\n";
    print $fh "#define $sendName \L${sendName}\n";
    print $fh "#define $recvName \L${recvName}\n";
    print $fh "#define $sendNameDebug \L${sendNameDebug}\n" if $opt_g;
    print $fh "#define $recvNameDebug \L${recvNameDebug}\n" if $opt_g;
    print $fh "\n#endif /* FM_NEED_UPPERCASE */\n";
    print $fh "\n#endif /* FM_NEED_TRAILING_UNDERSCORES */\n";
    
    &genSend(0, $sendName);
    &genRecv(0, $recvName);
    if ($opt_g)
    {
	&genSend(1, $sendNameDebug);
	&genRecv(1, $recvNameDebug);
    }
}

sub genSend
{
    local($withDebug, $sendName) = @_;
    local($i, $j);
    local(@decls, $nbufs, $offset);

    print $fh "void $sendName(f_integer *f_outport\n";

    #
    # Declarations for arguments
    #
    for ($i = 0; $i < $n_elts; $i++)
    {
	print $fh "\t, $elt_ctype[$i] $elt_star[$i]$elt_id[$i]";
	if ($elt_is_array[$i])
	{
	    print $fh ", f_integer *$elt_id[$i]_size";
	}
	print $fh "\n";
    }
    
    print $fh "\t, f_integer *iostat, f_integer *have_iostat\n";
    print $fh "\t, f_character file, f_integer *line\n" if $withDebug;
    print $fh $charLenDecls;
    print $fh "\tCHAR_LEN_DECL(file_len)" if $withDebug;
    print $fh ")\n";

    print $fh "{\n";

    #
    # Local variables and common code.
    #
    print $fh (<<END);

    fm_outport_t outport;
    yoke_buffer_t buf;
    conn_t dest;
    int buffer_size, count;
END
    print $fh "    LOCAL_CHAR_LEN_DECL(file_len)" if $withDebug;
    print $fh "    char file_name[1024];\n" if $withDebug;

    #
    # Local vars for port pointers
    #

    #
    # If there are no arrays of ports in this port, we can statically
    # allocate the port lists.
    #

    local($have_inport_arrays, $have_outport_arrays);
    local($inport_count, $outport_count);

    $inport_count = $outport_count = 0;

    for $i (@inport_elts)
    {
        if ($elt_is_array[$i])
        {
            $have_inport_arrays = 1;
            last;
        }
        else
        {
            $inport_count++;
        }
    }
        
    for $i (@outport_elts)
    {
        if ($elt_is_array[$i])
        {
            $have_outport_arrays = 1;
            last;
        }
        else
        {
            $outport_count++;
        }
    }

    print $fh "    int p_count, i;\n" if @port_elts;

    if (@inport_elts)
    {
        if ($have_inport_arrays)
        {
            print $fh "    struct _fm_inport_migrec_t *inports;\n";
        }
        else
        {
            print $fh "    struct _fm_inport_migrec_t inports[$inport_count];\n";
        }
	print $fh "    int n_inports = 0, inport_idx = 0;\n";
    }

    if (@outport_elts)
    {
        if ($have_outport_arrays)
        {
            print $fh "    struct _fm_outport_migrec_t *outports;\n";
        }
        else
        {
            print $fh "    struct _fm_outport_migrec_t outports[$outport_count];\n";
        }
	print $fh "    int n_outports = 0, outport_idx = 0;\n";
    }
        
    for $i (@port_elts)
    {
        print $fh "    $port_type[$i] *A_${i}_port;\n";
    }

    #
    # Decls for character args
    #
    print $fh $localCharVarDecls;

    print $fh "    StartSafeRegion;\n";
    print $fh "    CONVERT_CHAR_FTN_TO_C_INPLACE(file, file_len, file_name);\n"
	if $withDebug;
    print $fh (<<END);

    outport = (fm_outport_t) *f_outport;

    if (outport->local_inport != (fm_inport_t) NULL)
	_fm_ensure_inport_has_channel(outport->local_inport);

    dest = outport->channel_conn;

    if (dest == (conn_t) NULL)
END
    if ($withDebug)
    {
	print $fh "    SetIOSTATWithLine(FM_IOSTAT_OUTPORT_UNCONNECTED, file_name, *line);\n";
    }
    else
    {
	print $fh "    SetIOSTAT(FM_IOSTAT_OUTPORT_UNCONNECTED);\n";
    }
    #
    # Count the number of port elements if we have arrays
    #

    if ($have_outport_arrays)
    {
	for $i (@port_elts)
	{
	    print $fh (<<END);
    for (i = 0; i < $elt_array_len[$i]; i++)
        n_$elt_type[$i]s++;
END
        }
    }
    else
    {
	print $fh "    n_inports = $inport_count;\n" if @inport_elts;
	print $fh "    n_outports = $outport_count;\n" if @outport_elts;
    }

    #
    # Allocate port arrays
    #

    if ($have_inport_arrays)
    {
	print $fh "    inports = (struct _fm_inport_migrec_t *) malloc(sizeof(struct _fm_inport_migrec_t) * n_inports);\n";
	print $fh "    MallocCheck($sendName, inports, struct _fm_inport_migrec_t *);\n";
    }

    if ($have_outport_arrays)
    {
	print $fh "    outports = (struct _fm_outport_migrec_t *) malloc(sizeof(struct _fm_outport_migrec_t) * n_outports);\n";
	print $fh "    MallocCheck($sendName, outports, fm_outport_migrec_t);\n";
    }


    #
    # Initialize port pointers
    #

    for $i (@inport_elts)
    {
        print $fh (<<END);
    A_${i}_port = ($port_type[$i] *) A_${i};
    for (i = 0; i < $elt_array_len[$i]; i++)
    {
	Assert(inport_idx < n_inports);
	inports\[inport_idx++\].inport =  A_${i}_port\[i\];
    }
END
    }
    for $i (@outport_elts)
    {
        print $fh (<<END);
    A_${i}_port = ($port_type[$i] *) A_${i};
    for (i = 0; i < $elt_array_len[$i]; i++)
    {
	Assert(outport_idx < n_outports);
	outports\[outport_idx\].pending = FALSE;
	outports\[outport_idx\].id = -1;
	outports\[outport_idx\].new_conn = (conn_t) NULL;
	outports\[outport_idx\].old_conn = (conn_t) NULL;
        outports\[outport_idx++\].outport = A_${i}_port\[i\];
    }
END
    }

    print $fh "    Assert(inport_idx == n_inports);\n" if @inport_elts;
    print $fh "    Assert(outport_idx == n_outports);\n" if @outport_elts;

    #
    # Perform any pre-send notifications and channel creations that
    # need to be done for any of the ports being passed.
    #

    local($n_outports, $n_inports, $outports, $inports);

    if (@inport_elts)
    {
	$n_inports = "n_inports";
	$inports = "inports";
    }
    else
    {
	$inports = "(struct _fm_inport_migrec_t *) NULL";
	$n_inports = "0";
    }

    if (@outport_elts)
    {
	$n_outports = "n_outports";
	$outports = "outports";
    }
    else
    {
	$outports = "(struct _fm_outport_migrec_t *) NULL";
	$n_outports = "0";
    }


    if (@port_elts)
    {
	if ($withDebug)
	{
	    print $fh (<<END);
	_fm_do_presend_debug(dest,
			     $inports,
			     $n_inports,
			     $outports,
			     $n_outports,
			     file,
			     *line
			     CHAR_LEN(file_len)
			     );
END
	}
	else
	{
	    print $fh "    _fm_do_presend(dest, $inports, $n_inports, $outports, $n_outports);\n";
	}
    }
    

    #
    # Set up for computing buffer size
    #
    if (@array_elts)
    {
	print $fh "    buffer_size = yoke_sizeof_int(dest, $n_array_elts);\n";
	print $fh "    count = $n_array_elts;\n";
    }
    else
    {
	print $fh "    buffer_size = 0;\n";
	print $fh "    count = 0;\n";
    }

    #
    # Compute buffer size
    #

    for $i (@port_elts)
    {
        print $fh (<<END);

    for (i = 0; i < $elt_array_len[$i]; i++)
    {
	buffer_size += _fm_sizeof_$elt_type[$i](dest, A_${i}_port\[i\], &p_count);
	count += p_count;
    }
END
    }

    for ($i = 0; $i < $n_elts; $i++)
    {
        if (!$elt_is_port[$i])
        {
            print $fh "    buffer_size += yoke_sizeof_$elt_ctype[$i](dest, $elt_array_len[$i]);\n";
            print $fh "    count++;\n";
        }
    }

    print $fh "    buf = yoke_allocate_buffer(dest, buffer_size, count);\n";
    
    for ($i = 0; $i < $n_elts; $i++)
    {
	if ($elt_is_array[$i])
	{
	    print $fh "    yoke_put_int(buf, $elt_id[$i]_size, 1);\n";
        }

        if ($elt_is_port[$i])
        {
            print $fh (<<END);
    for (i = 0; i < $elt_array_len[$i]; i++)
        _fm_put_$elt_type[$i](buf, $elt_id[$i]_port\[i\]);
END
        }
        elsif ($charVar[$i])
	{
	    print $fh "    EXTRACT_CHAR_PTR($elt_id[$i], $charVar[$i])\n";
	    print $fh "    yoke_put_char(buf, $charVar[$i], $elt_array_len[$i]);\n";
	}
	else
        {
	    print $fh "    yoke_put_$elt_ctype[$i](buf, $elt_id[$i], $elt_array_len[$i]);\n";
	}
    }

    if ($withDebug)
    {
	print $fh (<<END);
    _fm_send_generic_debug(outport, buf, iostat, have_iostat, file, *line
			   CHAR_LEN(file_len));
END
    }
    else
    {
	print $fh "    _fm_send_generic(outport, buf, iostat, have_iostat);\n";
    }
    print $fh "    EndSafeRegion;\n";


    print $fh "}\n\n";
}

sub genRecv
{
    local($withDebug, $recvName) = @_;
    local($i, $j);
    local(@decls, $nbufs, $offset);
    local($type, $size);
    local($mdecl);

    print $fh "void $recvName(f_integer *f_inport\n";

    for ($i = 0; $i < $n_elts; $i++)
    {
	print $fh "\t, $elt_ctype[$i] $elt_star[$i]$elt_id[$i]\n";
    }
    
    print $fh "\t, f_integer *iostat, f_integer *have_iostat\n";
    print $fh "\t, f_character file, f_integer *line\n" if $withDebug;
    print $fh $charLenDecls;
    print $fh "\tCHAR_LEN_DECL(file_len)" if $withDebug;
    print $fh ")\n";

    print $fh "{\n";

    $mdecl = "int msgsize;\n" if @array_elts;
    print $fh $localCharVarDecls;

    print $fh (<<END);
    int i;
    fm_inport_t inport;
    yoke_buffer_t buf;
    LOCAL_CHAR_LEN_DECL(file_len)
    $mdecl
END

    #
    # Declarations and initialization for port variables
    #

    for $i (@inport_elts)
    {
	print $fh "    fm_inport_t *A_${i}_port = (fm_inport_t *) A_$i;\n";
    }

    for $i (@outport_elts)
    {
	print $fh "    fm_outport_t *A_${i}_port = (fm_outport_t *) A_$i;\n";
    }

    print $fh "    StartSafeRegion;\n";
    print $fh "    inport = (fm_inport_t) *f_inport;\n";

    if ($withDebug)
    {
	print $fh (<<END);
    buf = _fm_recv_generic_debug(inport, iostat, have_iostat, file, *line
				 CHAR_LEN(file_len));
END
    }
    else
    {
	print $fh "    buf = _fm_recv_generic(inport, iostat, have_iostat);\n";
    }


    print $fh "    if (buf)\n";
    print $fh "    {\n";

    for ($i = 0; $i < $n_elts; $i++)
    {
	if ($elt_is_array[$i])
	{
	    print $fh "        yoke_get_int(buf, &msgsize, 1);\n";
	    $size = "msgsize";
	}
	else
	{
	    $size = "1";
	}

	if ($elt_is_port[$i])
	{
	    print $fh (<<END);
        for (i = 0; i < $size; i++)
	{
	    _fm_invalidate_$elt_type[$i]($elt_id[$i]_port\[i\]);
	    _fm_get_$elt_type[$i](buf, &($elt_id[$i]_port\[i\]));
	    _fm_pindex_add_$elt_type[$i]($elt_id[$i]_port\[i\]);
	}
END
	}
        elsif ($charVar[$i])
	{
	    print $fh "        EXTRACT_CHAR_PTR($elt_id[$i], $charVar[$i])\n";
	    print $fh "        yoke_get_char(buf, $charVar[$i], $size);\n";
	}
	else
	{
	    print $fh "        yoke_get_$elt_ctype[$i](buf, $elt_id[$i], $size);\n";
	}
    }
    print $fh "        yoke_free_buffer(buf);\n";
    print $fh "    }\n";

    print $fh "    EndSafeRegion;\n";
    
    print $fh "}\n\n";
}


1;
