#
# Please see the COPYRIGHT_and_DISCLAIMER file in the top level
# directory of the distribution regarding the provisions under
# which this software is distributed.
#
#
# Transform PROCESSORS statements.
#
# If we are in the main program, replace the declaration with a a
# declaration of _processes(n) where n is the number of dimensions
# in the PROCESSORS declaration, and initialize the elements to the
# values of the declaration. 
#
# For non-PROGRAM subprograms, we need to declare a local processors array
# of suitable dimension, and initialize it via a FM runtime function
# which will look at the current virtual computer size/shape and initialize
# the array appropriately. The variables in the PROCESSORS declaration
# will be set to the values in the processors array. The local processors 
# array will be the one used by any annotations.
#


sub xformProcessors
{
    local($bounds, @dims, $i, $procs, $ndims);
    
    print "xformProcessors on $CurrentProgram\n" if $opt_v;

    local($flavor, $isParallel, $returnType) =
	split(/$;/, $Programs{$CurrentProgram});

    $bounds = &vardict'get("__machine", "bounds");
    
    if (!$bounds)
    {
	if ($flavor eq 'program')
	{
	    $bounds = '1';
	    print "Setting default bounds\n" if $opt_v;
	    &vardict'set("__machine", "bounds", $bounds);
	}
	else
	{
	    return;
	}
    }

    print "Program has bounds $bounds\n" if $opt_v;

    @dims = split(/\|/, $bounds);
    $ndims = @dims;
#    print "Dims are ", join(":", @dims), ", $dims\n";
    
    $procs = 'fm_processors';
    &declareVar('integer', $procs);
    &dimensionVar($procs, $ndims + 1);

    &addInitCode("$procs(1) = " . join(" * ", @dims));
    for ($i = 0; $i < $ndims; $i++)
    {
	&addInitCode("$procs(" . ($i+2) . ") = $dims[$i]");
    }

    &vardict'set('__processors_var', 'name', $procs);
    
    if ($flavor eq 'program')
    {
	&addInitCode("call fm_init($procs, " . ($ndims + 1) . ")");
	if ($opt_proc)
	{
	    &addCleanupCode("call fm_shutdown");
	}
	else
	{
	    &addCleanupCode("call fm_cleanup");
	}
    }
    else
    {
#	&addInitCode("call fm_verify_processors_decl($procs, " . ($ndims+1) . ")");
	&addInitCode("call fm_set_vc_size($procs, " . ($ndims+1) . ")");
    }
}

1;
