/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

/*
 * Macros for abstracted Fortran string handling.
 *
 * String arguments in a C routine that is Fortran-callable
 * should be writting as follows. Both the CHAR_LEN_DECL and the
 * LOCAL_CHAR_LEN_DECL are required.
 *
 * 	CFUNC(string_arg CHAR_LEN_DECL(string_len))
 *      {
 *          char *cstr, buf[1000];
 *	    LOCAL_CHAR_LEN_DECL(string_len)
 *          CONVERT_CHAR_FTN_TO_C_MALLOC(string_arg, string_len, cstr);
 *          CONVERT_CHAR_FTN_TO_C_INPLACE(string_arg, string_len, buf);
 *      }
 */
#ifdef CRAY
#include <fortran.h>
#define CHAR_LEN_DECL(varname)
#define CHAR_LEN(varname)
#define LOCAL_CHAR_LEN_DECL(varname) int varname;
#define CONVERT_CHAR_FTN_TO_C_INPLACE(ftn_char, len, buf) \
	{ int cpy_len; \
	len = _fcdlen(ftn_char); \
	cpy_len = (len >= sizeof(buf)) ? (sizeof(buf) - 1) : len; \
	strncpy(buf, _fcdtocp(ftn_char), cpy_len); \
	buf[cpy_len] = '\0'; \
        }
#define CONVERT_CHAR_FTN_TO_C_MALLOC(ftn_char, len, ptr) \
	len = _fcdlen(ftn_char); \
	ptr = (char *) malloc(len + 1); \
	MallocCheck(ConvertCharFtnToC(), ptr, char *); \
	strncpy(ptr, _fcdtocp(ftn_char), len); \
	ptr[len] = '\0';
#define CONVERT_CHAR_C_TO_FTN(ftn_char, buf, len) \
	len = strlen(buf); \
	ftn_char = _cptofcd(buf, len);
#define EXTRACT_CHAR_PTR(ftn_char, ptr) \
	ptr = _fcdtocp(ftn_char);
#define EXTRACT_CHAR_LEN(ftn_char, len) \
	len = _fcdlen(ftn_char);

#else

#define CHAR_LEN_DECL(varname) , int varname
#define CHAR_LEN(varname) , varname
#define LOCAL_CHAR_LEN_DECL(varname)
#define CONVERT_CHAR_FTN_TO_C_INPLACE(ftn_char, len, buf) \
	{ int cpy_len; \
	cpy_len = (len >= sizeof(buf)) ? (sizeof(buf) - 1) : len; \
	strncpy(buf, ftn_char, cpy_len); \
	buf[cpy_len] = '\0'; \
        }
#define CONVERT_CHAR_FTN_TO_C_MALLOC(ftn_char, len, ptr) \
	ptr = (char *) malloc(len + 1); \
	MallocCheck(ConvertCharFtnToC(), ptr, char *); \
	strncpy(ptr, ftn_char, len);\
	ptr[len] = '\0';
#define CONVERT_CHAR_C_TO_FTN(ftn_char, buf, len) \
	len = strlen(buf); \
	ftn_char = buf;
#define EXTRACT_CHAR_PTR(ftn_char, ptr) \
	ptr = ftn_char;
#define EXTRACT_CHAR_LEN(ftn_char, len) 
#endif

/*
 * Macros for dealing with logical types.
 */
#ifdef CRAY

#define FTN_LOGICAL_TRUE _btol(1)
#define FTN_LOGICAL_FALSE _btol(0)

/* Argument to FTN_LOGICAL_TO_LONG is a pointer */
#define FTN_LOGICAL_TO_LONG(v) _ltob(v)

/* Argument to FTN_LOGICAL_TO_LONG is an integer */
#define LONG_TO_FTN_LOGICAL(v) _btol(v)

#else

#define FTN_LOGICAL_TRUE 1
#define FTN_LOGICAL_FALSE 0

/* Argument to FTN_LOGICAL_TO_LONG is a pointer */
#define FTN_LOGICAL_TO_LONG(v) ((long) (*(v)))

/* Argument to FTN_LOGICAL_TO_LONG is an integer */
#define LONG_TO_FTN_LOGICAL(v) ((long) (v) == 0 ? 0 : 1)

#endif

/*
 * Fortran data types
 */

#ifdef CRAY

typedef _fcd f_character;
typedef _f_log f_logical;
typedef _f_int f_integer;
typedef _f_real f_real;
typedef _f_real16 f_doubleprecision;

#else

typedef char *f_character;
typedef int f_logical;
typedef int f_integer;
typedef float f_real;
typedef double f_doubleprecision;

#endif


typedef int f_in_port;
typedef int f_out_port;

typedef long f_address;

/*
 * Aliases for the yoke_put/yoke_get routines for the fortran types.
 */

#ifdef CRAY

#define yoke_put_f_character yoke_put_char
#define yoke_put_f_integer yoke_put_int
#define yoke_put_f_logical yoke_put_int
#define yoke_put_f_real yoke_put_double
#define yoke_put_f_doubleprecision yoke_put_long_double
#define yoke_put_f_in_port yoke_put_int
#define yoke_put_f_out_port yoke_put_int
#define yoke_put_f_address yoke_put_long

#define yoke_get_f_character yoke_get_char
#define yoke_get_f_integer yoke_get_int
#define yoke_get_f_logical yoke_get_int
#define yoke_get_f_real yoke_get_double
#define yoke_get_f_doubleprecision yoke_get_long_double
#define yoke_get_f_in_port yoke_get_int
#define yoke_get_f_out_port yoke_get_int
#define yoke_get_f_address yoke_get_long

#define yoke_sizeof_f_character yoke_sizeof_char
#define yoke_sizeof_f_integer yoke_sizeof_int
#define yoke_sizeof_f_logical yoke_sizeof_int
#define yoke_sizeof_f_real yoke_sizeof_double
#define yoke_sizeof_f_doubleprecision yoke_sizeof_long_double
#define yoke_sizeof_f_in_port yoke_sizeof_int
#define yoke_sizeof_f_out_port yoke_sizeof_int
#define yoke_sizeof_f_address yoke_sizeof_long

#endif /* CRAY */

#ifndef yoke_put_f_character

#define yoke_put_f_character yoke_put_char
#define yoke_put_f_integer yoke_put_int
#define yoke_put_f_logical yoke_put_int
#define yoke_put_f_real yoke_put_float
#define yoke_put_f_doubleprecision yoke_put_double
#define yoke_put_f_in_port yoke_put_int
#define yoke_put_f_out_port yoke_put_int
#define yoke_put_f_address yoke_put_long

#define yoke_get_f_character yoke_get_char
#define yoke_get_f_integer yoke_get_int
#define yoke_get_f_logical yoke_get_int
#define yoke_get_f_real yoke_get_float
#define yoke_get_f_doubleprecision yoke_get_double
#define yoke_get_f_in_port yoke_get_int
#define yoke_get_f_out_port yoke_get_int
#define yoke_get_f_address yoke_get_long

#define yoke_sizeof_f_character yoke_sizeof_char
#define yoke_sizeof_f_integer yoke_sizeof_int
#define yoke_sizeof_f_logical yoke_sizeof_int
#define yoke_sizeof_f_real yoke_sizeof_float
#define yoke_sizeof_f_doubleprecision yoke_sizeof_double
#define yoke_sizeof_f_in_port yoke_sizeof_int
#define yoke_sizeof_f_out_port yoke_sizeof_int
#define yoke_sizeof_f_address yoke_sizeof_long

#endif /* notdef yoke_sizeof... */
