/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */

typedef struct _fm_inport_t *fm_inport_t;

/*
 * An outport is defined if either it has a channel identifier
 * (meaning it is channeled to a remote inport) or it has a
 * local_inport pointer.
 */
typedef struct _fm_outport_t
{
#ifdef DEBUG
    int start_magic;
#endif
    int channel_id;
    conn_t channel_conn;
    fm_inport_t local_inport;
    conn_t future_home;
    int inport_index;
    bool_t closed;
    bool_t passed_in_call;
#ifdef DEBUG
    char *name;
    int end_magic;
#endif
} *fm_outport_t;

typedef struct _fm_outport_reference_t
{
    fm_outport_t local_outport;
    conn_t remote_outport;
} fm_outport_reference_t;


/*
 * An inport is defined if it either has a nonzero outport count.
 */
struct _fm_inport_t
{
#ifdef DEBUG
    int start_magic;
#endif
    int channel_id;
    conn_t channel_conn;
    
    fm_outport_reference_t *outports;
    int n_outports, n_outports_alloc;
    bool_t passed_in_call;
#ifdef DEBUG
    char *name;
    int end_magic;
#endif
};

typedef void (*fm_copy_func_t)(void *args, yoke_buffer_t buf);


typedef struct _fm_inport_migrec_t
{
    fm_inport_t inport;
    conn_t old_conn;
    int old_id;
    conn_t new_conn;
    int new_id;
    bool_t pending;
} *fm_inport_migrec_t;

typedef struct _fm_outport_migrec_t
{
    fm_outport_t outport;
    int id;
    conn_t old_conn, new_conn;
    bool_t pending;
} *fm_outport_migrec_t;

typedef struct _fm_process_block_entry_t
{
    conn_t process;
    yoke_buffer_t return_value;
    void *out_arg_list;
    fm_copy_func_t copy_func;

    int n_inport_migrations;
    int n_outport_migrations;

    fm_inport_migrec_t inport_migrecs;
    fm_outport_migrec_t outport_migrecs;

    fm_inport_t *inports;
    int n_inports;

    fm_outport_t *outports;
    int n_outports;
} fm_process_block_entry_t;

typedef struct _fm_process_block_t
{
    char subprogram_name[128];
    int block_number;
    int size;
    int current_slot;
    int *processors;
    int n_processor_dims;
    int n_migrations;
    int n_acks_wanted;
    fm_process_block_entry_t *entries;
} *fm_process_block_t;

typedef void (*fm_remote_start_func_t)(conn_t, yoke_buffer_t);
typedef int fm_remote_start_func_id_t;

typedef struct _fm_remote_start_func_registry_t
{
    fm_remote_start_func_id_t *id;
    fm_remote_start_func_t func;
    char *name;
    char *process;
} fm_remote_start_func_registry_t;

typedef struct _fm_pausepoint_t
{
    int id;
    bool_t pause_me;
    int sleep_me;
    struct _fm_pausepoint_t *next;
} *fm_pausepoint_t;


/*
 * This is used in migrate.c to handle enqueuing
 * of migration requests that arrive before a process
 * gets started.
 */
typedef struct _fm_migration_queue_entry_t
{
    int type;
    yoke_buffer_t buf;
    struct _fm_migration_queue_entry_t *next;
} *fm_migration_queue_entry_t;
