/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */


#include "types.h"

/* A parse tree node */

typedef union _leaf_data
{
    int idata;
    char *sdata;
    double ddata;
} leaf_data;

struct _tree_node_t;

typedef struct _tree_node_list
{
    struct _tree_node **nodes;
    int nnodes;
    int nnodes_alloc;
} tree_node_list;

typedef struct _tree_node
{
    enum node_type type;
    leaf_data data;
    int line;
    char *file;
    int label;
    tree_node_list *kids;
} tree_node;

typedef struct _tree_leaf_node
{
    char type;
    leaf_data data;
    int line;
    char *file;
    int label;

} tree_leaf_node;

#ifdef __STDC__
void fatal(char *, ...);
#else
void fatal();
#endif

#ifdef PARANOID

int paranoid_has_kids();
int paranoid_nkids();
tree_node *paranoid_kid();

int paranoid_is_tree_node();

#define PARANOID_TREE_NODE_CHECK(t) \
    if (!paranoid_is_tree_node(t)) fatal("PARANOID ERROR: %x not a tree node", t);

#define HAS_KIDS(t) paranoid_has_kids(t)
#define NKIDS(t) paranoid_nkids(t)
#define KID(t,n) paranoid_kid(t,n)
#define SET_KID(t,n,k) paranoid_set_kid(t,n,k)
#define REAL_HAS_KIDS(t) ((t)->kids != NULL && (t)->kids->nnodes > 0)
#define REAL_NKIDS(t) ((t)->kids == NULL ? 0 : (t)->kids->nnodes)
#define REAL_KID(t,n) ((t)->kids == NULL ? (tree_node *) NULL : (t)->kids->nodes[n])
#define REAL_SET_KID(t,n,k) {if ((t)->kids != NULL) { (t)->kids->nodes[n] = k; }}
#else
#define HAS_KIDS(t) ((t)->kids != NULL && (t)->kids->nnodes > 0)
#define NKIDS(t) ((t)->kids == NULL ? 0 : (t)->kids->nnodes)
#define KID(t,n) ((t)->kids == NULL ? (tree_node *) NULL : (t)->kids->nodes[n])
#define SET_KID(t,n,k) {if ((t)->kids != NULL) { (t)->kids->nodes[n] = k; }}
#endif

tree_node *new_internal_node();
tree_node *new_int_node();
tree_node *new_string_node();
tree_node *new_string_node_len();
tree_node *new_double_node();
tree_node *new_nil_node();

tree_node *new_node();

#ifdef __STDC__
tree_node *new_nodev(int, int, ...);
#else
tree_node *new_nodev();
#endif

tree_node *map_list_0();
tree_node *map_list_1();

tree_node *add_kid();
tree_node *add_kid_to_front();
tree_node *add_kid2();
tree_node *add_kid3();
tree_node *add_kid4();
tree_node * add_int_kid();
tree_node * add_string_kid();
tree_node * add_double_kid();
tree_node * add_nil_kid();

int node_equal();

void print_node();
void pprint_node();

#ifdef GLOBALS
#define GLOBAL
#else /* GLOBALS */
#define GLOBAL extern
#endif

GLOBAL char *current_file;
GLOBAL int current_line;
