/*
 * Please see the COPYRIGHT_and_DISCLAIMER file in the top level
 * directory of the distribution regarding the provisions under
 * which this software is distributed.
 */


/*
 * This assume 4 byte unsigned long's, and a single byte character.
 *
 * Cray note: To have the Cray interoperate with other machine, we
 * need to pack into the same 9 byte header.  But in a Cray only
 * version it might be more efficient to just send 17 byte headers.
 *
 * The TCP_MSG_HEADER_MALLOC_SIZE, is the amount of space that
 * will be malloced for the header in a message buffer.  This may
 * be different than the TCP_MSG_HEADER_SIZE if you want to pad
 * the header out to, say, a word boundary.
 */
#ifdef _CRAY

#define TCP_MSG_HEADER_SIZE		17
#define TCP_MSG_HEADER_MALLOC_SIZE	24

#else  /* CRAY */

#define TCP_MSG_HEADER_SIZE		9
#define TCP_MSG_HEADER_MALLOC_SIZE	12

#endif /* CRAY */


typedef struct _tcp_buffer_t
{
#ifdef DEBUG
    int magic;
#endif
    yoke_proto_buffer_funcs_t funcs;
    struct _tcp_buffer_t *next;
    conn_t dest;
    unsigned long size;
    unsigned long type;
    int n_items;
    char *header_pointer;
    char *base_pointer;
    char *current_pointer;
} tcp_buffer_t;


typedef struct _tcp_proto_data_t
{
    proto_type_t type;
    char *host;
    int port;
    int fd;
    conn_t conn;
    int reference_count;
    struct _tcp_proto_data_t *next, *prev;

    /*
     * The rest of this is for handling non-blocking reads and writes,
     * and the resulting partial messages that this data structure must
     * hang onto.
     */
    int   recv_state;		/* receiving header or body */
    tcp_buffer_t *recv_buf;	/* Start of buffer we're receiving into */
    char *recv_buf_current;	/* Place in buffer that we're receving into */
    int   recv_n_left;		/* # bytes left to receive */
    tcp_buffer_t *head;	/* head of message queue for this proto */
    tcp_buffer_t *tail;	/* tail of queue */
    char header_buf[TCP_MSG_HEADER_SIZE]; /* space to recv message header */
} tcp_proto_data_t;


void			_y_tcp_process_arguments_init();
extern int 		_y_tcp_process_arguments(int curr_arg, int arg_count);
extern void		_y_tcp_usage_message();
extern yoke_proto_t 	_y_tcp_init();
extern conn_t 		_y_tcp_start_nodes(int *n_nodes, int *my_node_number);
extern conn_t 		_y_tcp_fork(int *child_pid);
